/*
 *
  ***** BEGIN LICENSE BLOCK *****
 
  Copyright (C) 2009-2019 Olof Hagsand and Benny Holmgren

  This file is part of CLIXON.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  Alternatively, the contents of this file may be used under the terms of
  the GNU General Public License Version 3 or later (the "GPL"),
  in which case the provisions of the GPL are applicable instead
  of those above. If you wish to allow use of your version of this file only
  under the terms of the GPL, and not to allow others to
  use your version of this file under the terms of Apache License version 2, 
  indicate your decision by deleting the provisions above and replace them with
  the  notice and other provisions required by the GPL. If you do not delete
  the provisions above, a recipient may use your version of this file under
  the terms of any one of the Apache License version 2 or the GPL.

  ***** END LICENSE BLOCK *****

 *
 *  Code for handling netconf hello messages
 *****************************************************************************/
/*
 Capabilities are advertised in messages sent by each peer during
   session establishment.  When the NETCONF session is opened, each peer
   (both client and server) MUST send a <hello> element containing a
   list of that peer's capabilities.  Each peer MUST send at least the
   base NETCONF capability, "urn:ietf:params:netconf:base:1.0".
<hello> 
    <capabilities> 
        <capability>URI</capability> 
    </capabilities> 
    
</hello> 
 */


#ifdef HAVE_CONFIG_H
#include "clixon_config.h" /* generated by config & autoconf */
#endif

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <assert.h>

/* cligen */
#include <cligen/cligen.h>

/* clicon */
#include <clixon/clixon.h>

#include "netconf_lib.h"
#include "netconf_hello.h"

static int
netconf_hello(cxobj *xn)
{
#ifdef nyi
    cxobj *x;

    x = NULL;

    while ((x = xpath_each(xn, "//capability", x)) != NULL) {
	
    }
#endif
    return 0;
}

int
netconf_hello_dispatch(cxobj *xn)
{
    cxobj *xp;
    int retval = -1;

    if ((xp = xpath_first(xn, "//hello")) != NULL)
	retval = netconf_hello(xp);
    return retval;
}

/*! Create Netconf hello. Single cap and defer individual to querying modules

 * @param[in]  h           Clicon handle
 * @param[in]  cb          Msg buffer
 * @param[in]  session_id  Id of client session
 * Lots of dependencies here. regarding the hello protocol.
 * RFC6241 NETCONF Protocol says: (8.1)
 *    MUST send a <hello> element containing a list of that peer's capabilities
 *    MUST send at least the base NETCONF capability, urn:ietf:params:netconf:base:1.1
 *    MAY include capabilities for previous NETCONF versions
 *    MUST include a <session-id>
 * the example shows urn:ietf:params:netconf:capability:startup:1.0 

 * RFC5277 NETCONF Event Notifications
 *  urn:ietf:params:netconf:capability:notification:1.0 is advertised during the capability exchange
 *  
 * RFC6022 YANG Module for NETCONF Monitoring
 *     MUST advertise the capability URI "urn:ietf:params:xml:ns:yang:ietf-netconf-monitoring"
 * RFC7895 Yang module library defines how to announce module features (not hell capabilities)
 * RFC7950 YANG 1.1 says (5.6.4);
 *    MUST announce the modules it implements by implementing the YANG module 
 *    "ietf-yang-library" (RFC7895) and listing all implemented modules in the
 *    "/modules-state/module" list.
 *    MUST advertise urn:ietf:params:netconf:capability:yang-library:1.0?
 *    revision=<date>&module-set-id=<id> in the <hello> message.
 *
 * Question: should the NETCONF in RFC6241 sections 8.2-8.9 be announced both 
 * as features and as capabilities in the <hello> message according to RFC6241?
 *   urn:ietf:params:netconf:capability:candidate:1:0 (8.3)
 *   urn:ietf:params:netconf:capability:validate:1.1 (8.6)
 *   urn:ietf:params:netconf:capability:startup:1.0 (8.7)
 *   urn:ietf:params:netconf:capability:xpath:1.0 (8.9)
 *   urn:ietf:params:netconf:capability:notification:1.0 (RFC5277)
 *
 * @note the hello message is created bythe netconf application, not the 
 *  backend, and backend may implement more modules - please consider if using
 *  library routines for detecting capabilities here. In contrast, yang module 
 * list (RFC7895) is processed by the backend.
 * @note encode bodies, see xml_chardata_encode()
 * @see yang_modules_state_get
 * @see netconf_module_load
 */
int
netconf_create_hello(clicon_handle h,
		     cbuf         *cb,
		     int           session_id)
{
    int   retval = -1;
    char *module_set_id;
    char *ietf_yang_library_revision;
    char *encstr = NULL;

    module_set_id = clicon_option_str(h, "CLICON_MODULE_SET_ID");
    if ((ietf_yang_library_revision = yang_modules_revision(h)) == NULL)
	goto done;
    add_preamble(cb);
    cprintf(cb, "<hello xmlns=\"%s\">", NETCONF_BASE_NAMESPACE);
    cprintf(cb, "<capabilities>");
        cprintf(cb, "<capability>urn:ietf:params:netconf:base:1.0</capability>");
    if (xml_chardata_encode(&encstr, "urn:ietf:params:netconf:capability:yang-library:1.0?revision=%s&module-set-id=%s",
			    ietf_yang_library_revision,
			    module_set_id) < 0)
	goto done;
    cprintf(cb, "<capability>%s</capability>", encstr);
    cprintf(cb, "<capability>urn:ietf:params:netconf:capability:candidate:1:0</capability>");
    cprintf(cb, "<capability>urn:ietf:params:netconf:capability:validate:1.1</capability>");
    cprintf(cb, "<capability>urn:ietf:params:netconf:capability:startup:1.0</capability>");
    cprintf(cb, "<capability>urn:ietf:params:netconf:capability:xpath:1.0</capability>");
    cprintf(cb, "<capability>urn:ietf:params:netconf:capability:notification:1.0</capability>");
    cprintf(cb, "</capabilities>");
    cprintf(cb, "<session-id>%lu</session-id>", (long unsigned int)session_id);
    cprintf(cb, "</hello>");
    add_postamble(cb);
    retval = 0;
 done:
    if (encstr)
	free(encstr);
    return retval;
}
